@@ -1,3 +1,4 @@
+import abc
 import collections
 import logging
 import base64
@@ -5,6 +6,7 @@ import json
 import os
 
 import six
+from six.moves import urllib
 
 
 log = logging.getLogger(__name__)
@@ -49,21 +51,31 @@ class Check(object):
         return {'script': script, 'interval': interval}
 
     @classmethod
-    def http(klass, url, interval, timeout=None):
+    def http(klass, url, interval, timeout=None, deregister=None, header=None):
         """
         Peform a HTTP GET against *url* every *interval* (e.g. "10s") to peform
-        health check with an option *timeout*
+        health check with an optional *timeout* and optional *deregister* after
+        which a failing service will be automatically deregistered. Optional
+        parameter *header* specifies headers sent in HTTP request. *header*
+        paramater is in form of map of lists of strings,
+        e.g. {"x-foo": ["bar", "baz"]}.
         """
         ret = {'http': url, 'interval': interval}
         if timeout:
             ret['timeout'] = timeout
+        if deregister:
+            ret['DeregisterCriticalServiceAfter'] = deregister
+        if header:
+            ret['header'] = header
         return ret
 
     @classmethod
-    def tcp(klass, host, port, interval, timeout=None):
+    def tcp(klass, host, port, interval, timeout=None, deregister=None):
         """
         Attempt to establish a tcp connection to the specified *host* and
-        *port* at a specified *interval* with optional *timeout*
+        *port* at a specified *interval* with optional *timeout* and optional
+        *deregister* after which a failing service will be automatically
+        deregistered.
         """
         ret = {
             'tcp': '{host:s}:{port:d}'.format(host=host, port=port),
@@ -71,6 +83,8 @@ class Check(object):
         }
         if timeout:
             ret['timeout'] = timeout
+        if deregister:
+            ret['DeregisterCriticalServiceAfter'] = deregister
         return ret
 
     @classmethod
@@ -82,11 +96,12 @@ class Check(object):
         return {'ttl': ttl}
 
     @classmethod
-    def docker(klass, container_id, shell, script, interval):
+    def docker(klass, container_id, shell, script, interval, deregister=None):
         """
         Invoke *script* packaged within a running docker container with
-        *container_id* at a specified specified *interval* on the configured
-        *shell* using the Docker Exec API
+        *container_id* at a specified *interval* on the configured
+        *shell* using the Docker Exec API.  Optional *register* after which a
+        failing service will be automatically deregistered.
         """
         ret = {
             'docker_container_id': container_id,
@@ -94,6 +109,8 @@ class Check(object):
             'script': script,
             'interval': interval
         }
+        if deregister:
+            ret['DeregisterCriticalServiceAfter'] = deregister
         return ret
 
     @classmethod
@@ -103,7 +120,8 @@ class Check(object):
             interval=None,
             ttl=None,
             http=None,
-            timeout=None):
+            timeout=None,
+            deregister=None):
 
         if not script and not http and not ttl:
             return {}
@@ -126,6 +144,9 @@ class Check(object):
             assert http
             ret['check']['timeout'] = timeout
 
+        if deregister:
+            ret['check']['DeregisterCriticalServiceAfter'] = deregister
+
         return ret
 
 
@@ -208,6 +229,39 @@ class CB(object):
         return cb
 
 
+class HTTPClient(six.with_metaclass(abc.ABCMeta, object)):
+    def __init__(self, host='127.0.0.1', port=8500, scheme='http',
+                 verify=True, cert=None):
+        self.host = host
+        self.port = port
+        self.scheme = scheme
+        self.verify = verify
+        self.base_uri = '%s://%s:%s' % (self.scheme, self.host, self.port)
+        self.cert = cert
+
+    def uri(self, path, params=None):
+        uri = self.base_uri + urllib.parse.quote(path, safe='/:')
+        if params:
+            uri = '%s?%s' % (uri, urllib.parse.urlencode(params))
+        return uri
+
+    @abc.abstractmethod
+    def get(self, callback, path, params=None):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def put(self, callback, path, params=None, data=''):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def delete(self, callback, path, params=None):
+        raise NotImplementedError
+
+    @abc.abstractmethod
+    def post(self, callback, path, params=None, data=''):
+        raise NotImplementedError
+
+
 class Consul(object):
     def __init__(
             self,
@@ -217,7 +271,8 @@ class Consul(object):
             scheme='http',
             consistency='default',
             dc=None,
-            verify=True):
+            verify=True,
+            cert=None):
         """
         *token* is an optional `ACL token`_. If supplied it will be used by
         default for all requests made with this client session. It's still
@@ -233,19 +288,26 @@ class Consul(object):
         By default the datacenter of the host is used.
 
         *verify* is whether to verify the SSL certificate for HTTPS requests
+
+        *cert* client side certificates for HTTPS requests
         """
 
         # TODO: Status
 
         if os.getenv('CONSUL_HTTP_ADDR'):
-            host, port = os.getenv('CONSUL_HTTP_ADDR').split(':')
+            try:
+                host, port = os.getenv('CONSUL_HTTP_ADDR').split(':')
+            except ValueError:
+                raise ConsulException('CONSUL_HTTP_ADDR (%s) invalid, '
+                                      'does not match <host>:<port>'
+                                      % os.getenv('CONSUL_HTTP_ADDR'))
         use_ssl = os.getenv('CONSUL_HTTP_SSL')
         if use_ssl is not None:
             scheme = 'https' if use_ssl == 'true' else 'http'
         if os.getenv('CONSUL_HTTP_SSL_VERIFY') is not None:
             verify = os.getenv('CONSUL_HTTP_SSL_VERIFY') == 'true'
 
-        self.http = self.connect(host, port, scheme, verify)
+        self.http = self.connect(host, port, scheme, verify, cert)
         self.token = os.getenv('CONSUL_HTTP_TOKEN', token)
         self.scheme = scheme
         self.dc = dc
@@ -255,6 +317,7 @@ class Consul(object):
 
         self.event = Consul.Event(self)
         self.kv = Consul.KV(self)
+        self.txn = Consul.Txn(self)
         self.agent = Consul.Agent(self)
         self.catalog = Consul.Catalog(self)
         self.health = Consul.Health(self)
@@ -288,7 +351,8 @@ class Consul(object):
                 body="",
                 node=None,
                 service=None,
-                tag=None):
+                tag=None,
+                token=None):
             """
             Sends an event to Consul's gossip protocol.
 
@@ -306,6 +370,10 @@ class Consul(object):
             *node*, *service*, and *tag* are regular expressions which remote
             agents will filter against to determine if they should store the
             event
+
+            *token* is an optional `ACL token`_ to apply to this request. If
+            the token's policy is not allowed to fire an event of this *name*
+            an *ACLPermissionDenied* exception will be raised.
             """
             assert not name.startswith('/'), \
                 'keys should not start with a forward slash'
@@ -316,6 +384,9 @@ class Consul(object):
                 params['service'] = service
             if tag is not None:
                 params['tag'] = tag
+            token = token or self.agent.token
+            if token:
+                params['token'] = token
 
             return self.agent.http.put(
                 CB.json(),
@@ -570,6 +641,39 @@ class Consul(object):
             return self.agent.http.delete(
                 CB.json(), '/v1/kv/%s' % key, params=params)
 
+    class Txn(object):
+        """
+        The Transactions endpoints manage updates or fetches of multiple keys
+        inside a single, atomic transaction.
+        """
+        def __init__(self, agent):
+            self.agent = agent
+
+        def put(self, payload):
+            """
+            Create a transaction by submitting a list of operations to apply to
+            the KV store inside of a transaction. If any operation fails, the
+            transaction is rolled back and none of the changes are applied.
+
+            *payload* is a list of operations where each operation is a `dict`
+            with a single key value pair, with the key specifying operation the
+            type. An example payload of operation type "KV" is
+            dict::
+
+                {
+                    "KV": {
+                      "Verb": "<verb>",
+                      "Key": "<key>",
+                      "Value": "<Base64-encoded blob of data>",
+                      "Flags": 0,
+                      "Index": 0,
+                      "Session": "<session id>"
+                    }
+                }
+            """
+            return self.agent.http.put(CB.json(), "/v1/txn",
+                                       data=json.dumps(payload))
+
     class Agent(object):
         """
         The Agent endpoints are used to interact with a local Consul agent.
@@ -1010,6 +1114,7 @@ class Consul(object):
             Returns *True* on success.
             """
             data = {'node': node, 'address': address}
+            params = {}
             dc = dc or self.agent.dc
             if dc:
                 data['datacenter'] = dc
@@ -1020,8 +1125,12 @@ class Consul(object):
             token = token or self.agent.token
             if token:
                 data['WriteRequest'] = {'Token': token}
+                params['token'] = token
             return self.agent.http.put(
-                CB.bool(), '/v1/catalog/register', data=json.dumps(data))
+                CB.bool(),
+                '/v1/catalog/register',
+                data=json.dumps(data),
+                params=params)
 
         def deregister(self,
                        node,
